@@ -20,12 +20,22 @@ class ProjectStatus(Enum):
     IN_PROGRESS = "IN_PROGRESS"
     COMPLETE = "COMPLETE"
     FAILED = "FAILED"
+    PENDING_RETRY = "PENDING_RETRY"
 
 
 class Project(pydantic.BaseModel):
     project_id: str
     status: ProjectStatus
     created_at: datetime
+    last_updated_at: datetime
+    # if started, represents % processing complete (out of 100)
+    percent_complete: int | None = None
+
+    def __str__(self) -> str:
+        ret = f"Project {self.project_id}\nstatus: {self.status.name}\ncreated at: {self.created_at}\nlast updated at: {self.last_updated_at}"
+        if self.percent_complete is not None:
+            ret += f"\npercent complete: {self.percent_complete}"
+        return ret
 
     @classmethod
     async def from_id(cls, project_id: str) -> "Project":
@@ -33,6 +43,7 @@ class Project(pydantic.BaseModel):
             project_id=project_id,
             status=ProjectStatus.NOT_STARTED,
             created_at=datetime.now(),
+            last_updated_at=datetime.now(),
         )
         await project.refresh()
         return project
@@ -59,8 +70,8 @@ class Project(pydantic.BaseModel):
 
         file_paths = local_file_utils.normalize_and_dedupe_paths(context_file_paths)
         project_root = None
-        if validate_lean_project_root:
-            example_file = file_paths[0] if file_paths else Path.cwd()
+        if validate_lean_project_root and file_paths:
+            example_file = file_paths[0]
             project_root = local_file_utils.find_lean_project_root(example_file)
 
         local_file_utils.validate_local_file_paths(
@@ -83,11 +94,12 @@ class Project(pydantic.BaseModel):
         batch_size: int = MAX_FILES_PER_REQUEST,
         validate_lean_project_root: bool = True,
     ) -> None:
+        assert len(context_file_paths) > 0, "No context files provided"
         file_paths = local_file_utils.normalize_and_dedupe_paths(context_file_paths)
 
         project_root = None
         if validate_lean_project_root:
-            example_file = file_paths[0] if file_paths else Path.cwd()
+            example_file = file_paths[0]
             project_root = local_file_utils.find_lean_project_root(example_file)
 
         local_file_utils.validate_local_file_paths(
@@ -165,15 +177,15 @@ class Project(pydantic.BaseModel):
                 file_path = Path(input_file_path)
                 file_content = local_file_utils.read_file_safely(file_path)
                 files = [("input_file", (str(file_path), file_content, "text/plain"))]
-                data = None
+                params = None
             else:
                 # Handle text input case
-                data = {"input_text": input_content}
+                params = {"input_text": input_content}
                 files = None
 
             response = await client.post(
                 f"/project/{self.project_id}/solve",
-                data=data,
+                params=params,
                 files=files,
             )
 
@@ -255,6 +267,7 @@ class Project(pydantic.BaseModel):
         validate_lean_project: Literal[True] = True,
         polling_interval_seconds: int = 30,
         max_polling_failures: int = 3,
+        output_file_path: Path | str | None = None,
     ) -> str: ...
 
     @overload
@@ -268,6 +281,7 @@ class Project(pydantic.BaseModel):
         validate_lean_project: bool = True,
         polling_interval_seconds: int = 30,
         max_polling_failures: int = 3,
+        output_file_path: Path | str | None = None,
     ) -> str: ...
 
     @classmethod
@@ -280,6 +294,7 @@ class Project(pydantic.BaseModel):
         wait_for_completion: bool = True,
         polling_interval_seconds: int = 30,
         max_polling_failures: int = 3,
+        output_file_path: Path | str | None = None,
     ) -> str:
         """Proves the input content.
 
@@ -292,6 +307,7 @@ class Project(pydantic.BaseModel):
             wait_for_completion: Whether to wait for the project to complete before returning. If False, the project id is returned.
             polling_interval_seconds: Interval in seconds to poll for the project status.
             max_polling_failures: Maximum number of polling failures before raising an error.
+            output_file_path: Path to save the solution file. If None, uses names the file based on the input file name with _aristotle appended.
 
         Returns:
             The file path to the solution, or the project id if wait_for_completion is False
@@ -324,11 +340,12 @@ class Project(pydantic.BaseModel):
                 context_file_paths = list(
                     local_file_utils.gather_file_imports(input_file_path, project_root)
                 )
-                await project.add_context(
-                    context_file_paths, validate_lean_project_root=True
-                )
+                if context_file_paths:
+                    await project.add_context(
+                        context_file_paths, validate_lean_project_root=True
+                    )
                 logger.info(f"Added {len(context_file_paths)} imports to project")
-            elif context_file_paths is not None:
+            elif context_file_paths is not None and len(context_file_paths) > 0:
                 logger.info("Adding context files to project...")
                 await project.add_context(
                     context_file_paths, validate_lean_project_root=validate_lean_project
@@ -346,8 +363,9 @@ class Project(pydantic.BaseModel):
             num_polling_failures = 0
             while project.status not in (ProjectStatus.COMPLETE, ProjectStatus.FAILED):
                 try:
+                    msg = str(project)
                     logger.info(
-                        f"Project status: {project.status} - sleeping for {polling_interval_seconds} seconds..."
+                        msg + f"\nSleeping for {polling_interval_seconds} seconds..."
                     )
                     await asyncio.sleep(polling_interval_seconds)
                     await project.refresh()
@@ -355,11 +373,11 @@ class Project(pydantic.BaseModel):
                     num_polling_failures += 1
                     if num_polling_failures >= max_polling_failures:
                         logger.error(
-                            "Too many errors polling for project status. Your project might still be running; try checking on it later with project id {project.project_id}."
+                            "Too many errors polling for project status. Your project might still be running; try checking on it yourself later with project id {project.project_id}."
                         )
                         raise
                     logger.warning(
-                        f"Error polling for project status. {num_polling_failures} failures so far. Sleeping for {polling_interval_seconds * num_polling_failures} seconds."
+                        f"We haven't been able to check on your project {num_polling_failures} time(s). Don't worry; we're still working on it. Trying again in {polling_interval_seconds * num_polling_failures} seconds."
                     )
                     await asyncio.sleep(polling_interval_seconds * num_polling_failures)
 
@@ -369,17 +387,20 @@ class Project(pydantic.BaseModel):
                 )
 
             logger.info("Solve complete! Getting solution...")
-            if "/" in str(input_file_path):
-                output_file_name = str(input_file_path).split("/")[-1]
-            else:
-                output_file_name = str(input_file_path).split("\\")[-1]
+            if output_file_path is None:
+                if "/" in str(input_file_path):
+                    output_file_path = str(input_file_path).split("/")[-1]
+                else:
+                    output_file_path = str(input_file_path).split("\\")[-1]
 
-            output_file_name = output_file_name.split(".")[0] + "_aristotle.lean"
-            solution_file_path = await project.get_solution(output_file_name)
+                output_file_path = output_file_path.split(".")[0] + "_aristotle.lean"
+            solution_file_path = await project.get_solution(
+                output_path=output_file_path
+            )
             logger.info(f"Solution saved to {solution_file_path}")
             return str(solution_file_path)
         finally:
-            if project.status in (ProjectStatus.QUEUED, ProjectStatus.IN_PROGRESS):
+            if project.status not in (ProjectStatus.FAILED, ProjectStatus.COMPLETE):
                 logger.info(
                     f"Project {project.project_id} is still running. You can manually check on it any time with Project.from_id('{project.project_id}')"
                 )
